/*********************************************************************NVMH3****

Copyright NVIDIA Corporation 2002
TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
*AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.


Comments:

******************************************************************************/

//
// demo.cpp
//
// Basic example of the use of the Cg runtime in a simple OpenGL
// program
//

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef WIN32
# include <windows.h>
#endif
#ifdef __APPLE__
# include <OpenGL/gl.h>
# include <OpenGL/glext.h>
# include <GLUT/glut.h>
#else
# include <GL/gl.h>
# include <GL/glut.h>
# include <GL/glext.h>
#endif
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#ifndef CWD
# define CWD ""
#endif


// Global variables: hold the Cg context that we're storing our programs
// in as well as handles to the vertex and fragment program used in this
// demo.

CGcontext context;
CGprogram vertexProgram, fragmentProgram, fragmentProgram_blur, fragmentProgram_laplacian, fragmentProgram_laplacian5;

char keypressed = 'b'; 

// Forward declarations of functions

static void keyboard(unsigned char key, int x, int y);
static void display();
static void handleCgError();
static void LoadTextures();
static void LoadCgPrograms();
static void ChooseProfiles();

///////////////////////////////////////////////////////////////////////////

// Main program; do basic GLUT and Cg setup, but leave most of the work
// to the display() function.

int main(int argc, char *argv[]) {
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(512, 512);
    glutCreateWindow("Cg demo");

    glutKeyboardFunc(keyboard);
    glutDisplayFunc(display);

    // Basic Cg setup; register a callback function for any errors
    // and create an initial context
    cgSetErrorCallback(handleCgError);
    context = cgCreateContext();

    // and all the rest happens in the display function...
    glutMainLoop();

    return 0;
}


// display callback function

static void display() 
{
    static int curTime = 0;

    static bool haveLoadedPrograms = false;
    if (!haveLoadedPrograms) {
	// Do one-time setup only once; setup Cg programs and textures
	// and set up OpenGL state.
        ChooseProfiles();
        LoadCgPrograms();
        LoadTextures();

        glEnable(GL_DEPTH_TEST);
        haveLoadedPrograms = true;
    }

    // The usual OpenGL stuff to clear the screen and set up viewing.
    glClearColor(.25, .25, .25, 1.);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	glOrtho(0, 1, 0, 1, .1, 20.); 

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(0, 0, 4, 0, 0, 0, 0, 1, 0);

    // Now make sure that the vertex and fragment programs, loaded
    // in LoadCgPrograms() are bound.
    cgGLBindProgram(vertexProgram);
	if(keypressed == 'b')
	{
		cgGLBindProgram(fragmentProgram_blur);
	}
	else if(keypressed ==  'e')
	{
		cgGLBindProgram(fragmentProgram);
	}
	else if(keypressed ==  'l')
	{
		cgGLBindProgram(fragmentProgram_laplacian);
	}
	else if(keypressed ==  'k')
	{
		cgGLBindProgram(fragmentProgram_laplacian5);
	}

    // Bind uniform parameters to vertex shader
    cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram, "ModelViewProj"),
                                CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                CG_GL_MATRIX_IDENTITY);
    cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram, "ModelView"),
                                CG_GL_MODELVIEW_MATRIX,
                                CG_GL_MATRIX_IDENTITY);
    cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram, "ModelViewIT"),
                                CG_GL_MODELVIEW_MATRIX,
                                CG_GL_MATRIX_INVERSE_TRANSPOSE);

    
    // And finally, enable the approprate texture for fragment shader; the
    // texture was originally set up in LoadTextures().
    cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram,
						   "diffuseMap"));

	cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram_blur,
						   "diffuseMap"));

	cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram_laplacian,
						   "diffuseMap"));
	cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram_laplacian5,
						   "diffuseMap"));
    // And go ahead and draw the scene geometry

	glBegin(GL_QUADS);

	glTexCoord3f(0, 0, 0);
	glVertex3f(0, 0, 0);

	glTexCoord3f(1, 0, 0);
	glVertex3f(1, 0, 0);

	glTexCoord3f(1, 1, 0);
	glVertex3f(1, 1, 0);

	glTexCoord3f(0, 1 , 0);
	glVertex3f(0, 1, 0);

	glEnd(); 
    
    // Disable the texture now that we're done with it.
    cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram,
						    "diffuseMap"));

	cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram_blur,
						    "diffuseMap"));

	cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram_laplacian,
					    "diffuseMap"));
	cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram_laplacian5,
					    "diffuseMap"));

    glutSwapBuffers();
       
    glutPostRedisplay();
}


// Choose the vertex and fragment profiles to use.  Try to use
// CG_PROFILE_ARBVFP1 and CG_PROFILE_ARBFP1, depending on hardware support.
// If those aren't available, fall back to CG_PROFILE_VP30 and
// CG_PROFILE_FP30, respectively.

CGprofile vertexProfile, fragmentProfile;

static void ChooseProfiles()
{
    // Make sure that the appropriate profiles are available on the
    // user's system.
    if (cgGLIsProfileSupported(CG_PROFILE_ARBVP1))
        vertexProfile = CG_PROFILE_ARBVP1;
    else {
        // try VP30
        if (cgGLIsProfileSupported(CG_PROFILE_VP30))
            vertexProfile = CG_PROFILE_VP30;
        else {
            fprintf(stderr, "Neither arbvp1 or vp30 vertex profiles supported on this system.\n");
            exit(1);
        }
    }

    if (cgGLIsProfileSupported(CG_PROFILE_ARBFP1))
        fragmentProfile = CG_PROFILE_ARBFP1;
    else {
        // try FP30
        if (cgGLIsProfileSupported(CG_PROFILE_FP30))
            fragmentProfile = CG_PROFILE_FP30;
        else {
            fprintf(stderr, "Neither arbfp1 or fp30 fragment profiles supported on this system.\n");
            exit(1);
        }
    }
}


static void LoadCgPrograms()
{
    assert(cgIsContext(context));

    // Load and compile the vertex program from demo_vert.cg; hold on to the
    // handle to it that is returned.
    vertexProgram = cgCreateProgramFromFile(context, CG_SOURCE, CWD "demo_vert.cg",
                                            vertexProfile, NULL, NULL);
    if (!cgIsProgramCompiled(vertexProgram))
        cgCompileProgram(vertexProgram);

    // Enable the appropriate vertex profile and load the vertex program.
    cgGLEnableProfile(vertexProfile);
    cgGLLoadProgram(vertexProgram);

    // And similarly set things up for the fragment program.
    fragmentProgram = cgCreateProgramFromFile(context, CG_SOURCE, CWD "demo_edge.cg",
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram))
        cgCompileProgram(fragmentProgram);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram);

	// And similarly set things up for the fragment program for blurring
    fragmentProgram_blur = cgCreateProgramFromFile(context, CG_SOURCE, CWD "demo_blur.cg",
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram_blur))
        cgCompileProgram(fragmentProgram_blur);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram_blur);

	// And similarly set things up for the fragment program for laplacian
    fragmentProgram_laplacian = cgCreateProgramFromFile(context, CG_SOURCE, CWD "demo_laplacian.cg",
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram_laplacian))
        cgCompileProgram(fragmentProgram_laplacian);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram_laplacian);

	// And similarly set things up for the fragment program for laplacian 5x5
    fragmentProgram_laplacian5 = cgCreateProgramFromFile(context, CG_SOURCE, CWD "demo_laplacian5x5.cg",
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram_laplacian5))
        cgCompileProgram(fragmentProgram_laplacian5);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram_laplacian5);


}

static void LoadTextures()
{
    // There is only one texture needed here--we'll read the input texture umbc.ppm here 

	FILE *fp; 
	char imageType[3],str[100];
	int xdim, ydim; 
	unsigned char *bufdata = new unsigned char[512*512*3], *bufp;

    GLuint handle;
    glGenTextures(1, &handle);

    // Basic OpenGL texture state setup
    glBindTexture(GL_TEXTURE_2D, handle);
    glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	
	fp = fopen("umbc.ppm", "rb");

	fgets(str,100,fp);
	sscanf(str,"%s",imageType);

	// skip comments embedded in header
    fgets(str,100,fp);
	while(str[0]=='#')
		fgets(str,100,fp);

	sscanf(str,"%d %d",&xdim, &ydim); 

	// read the next line into dummy variable
    fgets(str,100,fp);

	bufdata = new unsigned char [xdim * ydim * 3];
	bufp = bufdata; 
	fread(bufdata,sizeof(unsigned char),xdim * ydim *3,fp);

	fclose(fp); 

	// Fill in the texture map.
#define RES 512
    float *data = new float[RES*RES*4];
    float *dp = data;
    for (int i = 0; i < RES; ++i) {
        for (int j = 0; j < RES; ++j) {            				
				*dp++ = (float)(*bufdata++)/255.0f;
				*dp++ = (float)(*bufdata++)/255.0f;
				*dp++ = (float)(*bufdata++)/255.0f;			
				*dp++ = 1.0f;
        }
    }

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, RES, RES, 0, GL_RGBA, GL_FLOAT, data);
    delete[] data;

    // Tell Cg which texture handle should be associated with the sampler2D
    // parameter to the fragment shader.
    cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram, "diffuseMap"),
                            handle);
	cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram_blur, "diffuseMap"),
                            handle);
	cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram_laplacian, "diffuseMap"),
                            handle);
	cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram_laplacian5, "diffuseMap"),
                            handle);

}


static void keyboard(unsigned char key, int x, int y)
{
	
	switch(key)
	{
		case 'b':
			keypressed = 'b';
			printf("keypressed = %c\n", keypressed);
			break;

		case 'e':
			keypressed = 'e';
			printf("keypressed = %c\n", keypressed);
			break;

		case 'l':
			keypressed = 'l';
			printf("keypressed = %c\n", keypressed);
			break;

		case 'k':
			keypressed = 'k';
			printf("keypressed = %c\n", keypressed);
			break;

		case 'q': case 'Q':  case 27: 
			cgDestroyContext(context);
			exit(0);
    }

}

static void handleCgError() 
{
    fprintf(stderr, "Cg error: %s\n", cgGetErrorString(cgGetError()));
    exit(1);
}
